﻿using System;
using System.Collections.Generic;
using System.IO;
using System.IO.Compression;
using System.Reflection;
using System.Runtime.InteropServices;
using System.Threading;
using System.Xml;
using Livescribe.DesktopSDK.AFP;
using Livescribe.DesktopSDK.PenComm;
using Livescribe.DesktopSDK.PenComm.Interop;
using Livescribe.DesktopSDK.PenData;
using PenComm = Livescribe.DesktopSDK.PenComm;

namespace DataCapture
{
    class DataCapture
    {
        private static Smartpens smartpens;
        private const string DOC_DATA_FILE = "data.zip";
        private const string APP_NAME = "DataCapture";
        private const string PENCOMM_LOG = "DataCapture.log";
        private const string PENLET_NAME = "HWS";
        private const string DATA_FILE = "data.txt";
        private static bool foundPen = false;
        private static bool handlerRegistered = false;
        private static string tempDir;
        private static string STROKE_FILE = Environment.CurrentDirectory + "\\protocol.js";
        private static string PENLET_DATA_FILE;
        private static Dictionary<long, int[]> strokeTime = new Dictionary<long, int[]>();
        private static string dirpath = Environment.GetFolderPath(Environment.SpecialFolder.Desktop) + "\\Recolha";



        /// <summary>
        /// This sample gets data from the GetBoundingBoxes penlet and displays it on the console.  After displaying
        /// the the data last captured using the penlet we establish a data handler and wait for new data to show up.
        /// </summary>
        /// <param name="args"></param>
        static void Main(string[] args)
        {
            tempDir = Path.GetTempPath();

            DirectoryInfo mainDir = new DirectoryInfo(dirpath);
            if (!mainDir.Exists)
            {
                mainDir.Create();
            }

            System.Console.Write("Please connect your pen to the dock.");

            InitializePenCommLib();

            // Keep processing until user presses a key to terminate
            while (!Console.KeyAvailable)
            {
                if (!DataCapture.foundPen)
                {
                    // If we've never found a pen remind the user that we're waiting for one.
                    Console.Write(".");
                }
                Thread.Sleep(2000);
            }
        }

        /// <summary>
        /// Initialize PenComm library
        /// </summary>
        private static void InitializePenCommLib()
        {
            smartpens = new Smartpens(true, PENCOMM_LOG);

            // The PulsePens object keeps track of attach event handlers in "PulsePenAttachNormalEvent"
            // Create a new callback object for our PenAttachEvent method and register it by adding it 
            // to the list of attach event handlers.
            smartpens.SmartpenAttachNormalEvent += new Smartpens.SmartpenChangeCallback(PenAttachEvent);

            // Do the same for our detach event handler.
            smartpens.SmartpenDetachNormalEvent += new Smartpens.SmartpenChangeCallback(PenDetachEvent);

            // Search for any already docked pens
            smartpens.Find();
        }

        /// <summary>
        /// Register our data handler (NewData).  It will be invoked whenever new data appears for our penlet of 
        /// interest (PENLET_NAME).
        /// </summary>

        private static void RegisterForNewApplicationData()
        {
            if (!handlerRegistered)
            {
                try
                {
                    // Application registration information.
                    RegAppInfo appInfo = new RegAppInfo();
                    appInfo.autoStart = PenAPI.AutoStartTypes.Disabled;
                    appInfo.path = Assembly.GetExecutingAssembly().Location;
                    appInfo.description = APP_NAME;
                    appInfo.version = "1.0";

                    // Register Application (that's us)
                    AppHandle appHandle = smartpens.DesktopApplications.RegisterApp(appInfo);


                    // Register for a data callback so that every time the penlet PENLET_NAME has new data, it will 
                    // be passed to our callback (NewData).
                    //PenAPI.DataCallback dataCallback = new PenAPI.DataCallback(NewData);
                    //smartpens.DesktopApplications.RegisterDataCallback(appHandle, dataCallback, null);

                    // Register Penlet Information of the GetBoundingBoxes penlet
                    RegPackageInfo regPackageInfo = new RegPackageInfo();
                    regPackageInfo.appHandle = appHandle;
                    regPackageInfo.penId = PenComm.PenId.Zero();
                    regPackageInfo.uniqueName = PENLET_NAME;

                    // Register the GetBoundingBoxes penlet
                    PackageHandle packageHandle = smartpens.DesktopApplications.RegisterPackage(regPackageInfo);
                    handlerRegistered = true;
                }
                catch (Exception e)
                {
                    HandleError(e);
                }
            }
        }

        /*private static PenAPI.DataOpTypes NewData(PenDataCallbackInfo dataInfo, IntPtr userParam)
        {
            try
            {
                // Construct a file name that will be unique to a single stroke in which to save our data.
                // Note:  we need to save the data in a file in the event that the stroke is transferred to 
                // us in multiple invocations of the the callback.  This may happen as a stroke can be 
                // comprised of many co-ordinates.
                String filePath = "Pen" + dataInfo.penId + "_Penlet" + dataInfo.packageHandle + "_Time" + dataInfo.startRtc + ".zip";
                System.Console.WriteLine("    NewData: " + (dataInfo.pos+1) + "-" + (dataInfo.pos+dataInfo.bufSize) + " of " + dataInfo.total + " bytes, status = " + dataInfo.packetStatus);
                System.Console.WriteLine("    writing to:  " + filePath);
                filePath = tempDir + filePath;

                // Create or append to the file depending on whether we're starting with the first co-ordinate
                BinaryWriter bw = null;
                if (dataInfo.pos == 0)
                {
                    bw = new BinaryWriter(new FileStream(filePath, FileMode.Create));
                }
                else
                {
                    bw = new BinaryWriter(new FileStream(filePath, FileMode.Append));
                }

                byte[] buf = new byte[dataInfo.bufSize];
                Marshal.Copy(dataInfo.buffer, buf, 0, dataInfo.bufSize);
                bw.Write(buf);
                bw.Close();

                if (dataInfo.packetStatus != PenAPI.DataPacketStatus.MorePackets)
                {
                    PenletData penletData = new Container(filePath).PenletData;
                    //DisplayData(penletData);
                }
            }
            catch (Exception e)
            {
                HandleError(e);
            }

            return PenAPI.DataOpTypes.Continue;
        }*/

        /// <summary>
        /// Read and display penlet data.
        /// </summary>
        private static string[] DisplayData(PenletData penletData, Smartpen pen)
        {
            string[] protocol = null;
            try
            {
                // The penlet data file is a zip file which is structured like this:
                //      userdata\[Pen Serial]\[Penlet Name]\[InstanceId]\...
                //
                // PenItem corresponds to [Pen Serial]
                // AppItem corresponds to [Penlet Name]
                // InstanceItem corresponds to [InstanceId]
                // 
                // The PENLET_NAME penlet data file is named:
                //      userdata\[Pen Serial]\[Penlet Name]\[InstanceId]\data.txt
                foreach (PenItem penItem in penletData.PenItems)
                {

                    foreach (AppItem appItem in penItem)
                    {

                        foreach (InstanceItem instanceItem in appItem.InstanceItems)
                        {
                            protocol = PenDataFileReader(pen, protocol, instanceItem);
                        }
                    }

                }
                penletData.FileContext.CloseContext();


            }
            catch (Exception e)
            {
                HandleError(e);
            }
            return protocol;
        }

        private static string[] PenDataFileReader(Smartpen pen, string[] protocol, InstanceItem instanceItem)
        {
            ContainerStream containerStream = instanceItem.OpenStream(DATA_FILE);
            StreamReader streamReader = new StreamReader(containerStream);
            String line;

            long strokeStartTimeTemp = 0;
            long protocolStartTime = 0;

            while ((line = streamReader.ReadLine()) != null)
            {
                //System.Console.WriteLine(line);

                char[] del = new char[] { ':' };
                string[] timesLine;
                string strokeStartTime;

                if (line.StartsWith("--"))
                {
                    protocolStartTime = 0;
                }
                else
                {
                    if (line.Trim().StartsWith("#"))
                    {
                        if (line.StartsWith("##"))
                        {
                            protocol = line.Substring(2).Split(':');
                        }
                        else
                        {
                            strokeStartTime = line.Substring(1);
                            strokeStartTimeTemp = long.Parse(strokeStartTime);
                            if (protocolStartTime == 0)
                            {
                                if (protocol != null)
                                {
                                    XMLWriter(pen, protocol);
                                    protocol = null;
                                    strokeTime = new Dictionary<long, int[]>();
                                }
                                protocolStartTime = strokeStartTimeTemp;
                            }
                        }
                    }
                    else
                    {

                        timesLine = line.Split(del, StringSplitOptions.RemoveEmptyEntries);
                        int w = 0;
                        int[] timesArray = new int[timesLine.Length];

                        foreach (string j in timesLine)
                        {
                            timesArray[w++] = int.Parse(j);
                        }

                        if (!strokeTime.ContainsKey(strokeStartTimeTemp))
                            strokeTime.Add(strokeStartTimeTemp, timesArray);
                    }
                }
            }
            if (protocol != null)
                XMLWriter(pen, protocol);
            streamReader.Close();
            containerStream.Close();
            return protocol;
        }

        /// <summary>
        /// This is called when a pen is attached.  The first time a pen is attached or if the PENLET_NAME penlet was
        /// never installed and run we want to check for the penlet and process the data.  After that we register for 
        /// new data and only respond to new data events.  If we only wanted to respond to new data we could register
        /// our application during initialization and not even have a pen attach handler.
        /// </summary>
        /// <param name="pen">The attached pen</param>
        private static void PenAttachEvent(Smartpen pen)
        {
            DataCapture.foundPen = true;
            //TextWriter strokeFile = new StreamWriter(STROKE_FILE);
            //Console.WriteLine("\n\n--> Pen " + pen.PenSerial + " connected!\n\n");
            PENLET_DATA_FILE = dirpath + "\\" + pen.PenID.ToString() + "_penletData";
            if (!handlerRegistered)
            {
                PenletData penletData = null;
                try
                {
                    // Retrieve the data for the GetBoundingBox penlet and store it in a temporary file.  The data is
                    // in zip format and can be processed with any zip reader.

                    pen.DataGet(PENLET_NAME, 0, PENLET_DATA_FILE);

                    // From the data file we create a container and extract a PenletData object from it for easy processing
                    penletData = new Container(PENLET_DATA_FILE).PenletData;
                }
                catch (Exception e)
                {
                    HandleError(e);
                    return;
                }
                DisplayData(penletData, pen);
                RegisterForNewApplicationData();
            }
            DirectoryInfo dir = new DirectoryInfo(dirpath);
            /*foreach (FileInfo fi in dir.GetFiles())
            {
                CompressFile(fi);
            }*/

            ClearPen(pen);
            //Console.Clear();
            Console.WriteLine("\n\nConnect new pen or press ENTER key to exit.");

        }

        private static void ClearPen(Smartpen pen)
        {

            pen.DataDeleteStrokes("HandSpy", 0, 0);
        }


        private static void DecodeStrokes(Smartpen pen, ChangeItem item, ulong time, TextWriter strokeFile, String startTime, String endTime)
        {

            // We are going to extract strokes from the pen and store them in a temporary file named 
            // using the paper product's name.  
            string fileName = String.Format("{0}_{1}", item.Title, DOC_DATA_FILE);
            try
            {
                // The pen stores strokes in files on the pen's flash memory system.  The pen uses one file
                // for each paper product and the name of the file is the GUID of the paper product.  We
                // retrieve the stroke data and store it locally in a temporary file.
                pen.DataGet(item.Guid, 0, fileName);

                // From the data file we create a container and extract a document object from it for easy processing
                Document doc = new Container(fileName).Document;

                // Strokes are organized by pages.  Decode strokes for each page.
                foreach (PageChangeItem pageInfo in item.Pages)
                {

                    //strokeFile.WriteLine(String.Format("Copy: {0}, Page: {1}", pageInfo.Copy, pageInfo.Page));

                    // For each page we find the unique pattern object for it.
                    // the PatternPage is collection of strokes created by pens.
                    PatternPage patternPage = doc.GetPatternPage(pageInfo.Page, pageInfo.Copy);
                    // Each element of the PatternPage is SmartPen object which contains
                    // pen information (such as PenSerial) and strokes created by the pen
                    foreach (StrokeOwner penData in patternPage)
                    {

                        foreach (KeyValuePair<long, Stroke> keyPair in penData)
                        {
                            // Get time at which the stroke was created
                            long creationTime = keyPair.Key;


                            // A single stroke is comprised of potentially many points.  
                            // The pen records up to 70 points per second.
                            Stroke stroke = keyPair.Value;
                            Shape shape = stroke;
                            System.Drawing.Point[] points = shape.GetVertices();
                            int nPoints = 0;
                            long coordTime;
                            int[] timesIncrements = getTimeIncrements(creationTime, points.Length);
                            if (timesIncrements != null)
                            {
                                strokeFile.WriteLine(String.Format("<strokes start=\"{0}\" end=\"{1}\">", startTime, endTime));
                                foreach (System.Drawing.Point point in points)
                                {

                                    // Write the point of strokes to file.
                                    // Upper left corner of page is 0,0 with the y coordinate increasing down the page.
                                    // The unit is Anoto Unit (au), which are roughly 677 dpi.
                                    //strokeFile.Write(String.Format(" ({0},{1})", point.X, point.Y));

                                    coordTime = creationTime + timesIncrements[nPoints++];
                                    strokeFile.WriteLine(String.Format("<stroke Time=\"{0}\" X=\"{1}\" Y=\"{2}\"/>", coordTime, point.X, point.Y));

                                }
                                strokeFile.WriteLine("</strokes>");
                            }
                        }


                    }
                }
                doc.Close();
            }
            catch (Exception e)
            {
                HandleError(e);
            }
            finally
            {
                File.Delete(fileName);
            }
        }

        static int MAX_DELTA = 200;

        static private int[] getTimeIncrements(long creationTime, int nPoints)
        {
            int[] timesIncrements = null;

            for (int delta = 0; delta < MAX_DELTA; delta++)
            {
                if (strokeTime.TryGetValue(creationTime - delta, out timesIncrements) && timesIncrements.Length == nPoints)
                    return timesIncrements;
            }
            /*
            timesIncrements = new int[nPoints];
            for (int i = 0; i < nPoints; i++)
                    timesIncrements[i] = 0;*/
            return timesIncrements;
        }


        static private void PenDetachEvent(Smartpen pen)
        {
            foundPen = false;
            handlerRegistered = false;
        }

        static private StreamWriter getWriter(String code)
        {

            String d = dirpath + String.Format("\\{0}_{1}_{2}_{3}", System.DateTime.UtcNow.Day, System.DateTime.UtcNow.Month, System.DateTime.UtcNow.Year, System.DateTime.UtcNow.Hour);
            DirectoryInfo dir = new DirectoryInfo(d);
            if (!dir.Exists)
            {
                dir.Create();
                Console.WriteLine("\nFolder created on the Desktop - " + d);
            }
            int nFiles = dir.GetFiles().Length + 1;
            string fileName = d + "\\" + nFiles + "_" + code + ".xml";
            foreach (FileInfo f in dir.GetFiles())
            {
                if (f.FullName.CompareTo(fileName) == 0)
                    fileName = d + "\\" + nFiles + "_" + code + "NEW" + ".xml";
            }
            return new StreamWriter(fileName);
        }



        static private void XMLWriter(Smartpen pen, string[] protocol)
        {
            int pageNum = Convert.ToInt32(protocol[0]) + 1;

            TextWriter xmlFileWriter = getWriter(pageNum + "");
            xmlFileWriter.WriteLine(String.Format("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n<tns:protocol code=\"000\" xmlns:tns=\"protocol\" xmlns:xsi=\"http://www.fpce.up.pt:8080/PenSpyPrototype\" xsi:schemaLocation=\"protocol Protocol.xsd \">"));
            xmlFileWriter.WriteLine(String.Format("<header>\n<subject Age=\"0\" Handedness=\"0\" Name=\"0\" Sex=\"0\">\n<Birthdate Day=\"0\" Month=\"0\" Year=\"0\"/>\n</subject>"));
            xmlFileWriter.WriteLine(String.Format("<school Classroom=\"0\" Grade=\"0\" Name=\"0\"/>\n<layout Layout=\"A\" Page=\"{0}\"/>\n</header>", pageNum));
            ulong time = 0;
            SmartpenChangeList changeList = pen.ChangeList;
            // Update the change list (which started out empty) so it contains all the changes since time.
            changeList.Update(time);
            foreach (ChangeItem item in changeList.ChangeItems)
            {
                if (item.Title.Equals("HandSpy"))
                {
                    DecodeStrokes(pen, item, time, xmlFileWriter, protocol[1], protocol[2]);
                }
            }
            xmlFileWriter.WriteLine("</tns:protocol>");
            xmlFileWriter.Flush();
            xmlFileWriter.Close();

        }

        private static void HandleError(Exception ex)
        {
            System.Console.WriteLine("\nException: " + ex.Message, "Error");
            System.Console.WriteLine(ex.StackTrace);
        }

        public static void CompressFile(FileInfo fi)
        {
            // Get the stream of the source file.
            using (FileStream inFile = fi.OpenRead())
            {
                // Prevent compressing hidden and 
                // already compressed files.
                string name = fi.FullName.Substring(0, fi.FullName.Length - 4);
                if ((File.GetAttributes(fi.FullName) & FileAttributes.Hidden) != FileAttributes.Hidden & fi.Extension != ".gz")
                {
                    // Create the compressed file.
                    using (FileStream outFile =
                                File.Create(name + ".gz"))
                    {
                        using (GZipStream Compress =
                            new GZipStream(outFile,
                            CompressionMode.Compress))
                        {
                            // Copy the source file into 
                            // the compression stream.
                            CopyTo(inFile, Compress);
                        }
                    }
                }
            }
        }


        public static void CopyTo(Stream input, Stream output)
        {
            byte[] buffer = new byte[16 * 1024];
            int bytesRead;
            while ((bytesRead = input.Read(buffer, 0, buffer.Length)) > 0)
            {
                output.Write(buffer, 0, bytesRead);
            }
        }

        public static DateTime FromEpochToHours(long seconds)
        {
            var epoch = new DateTime(1970, 1, 1, 0, 0, 0, DateTimeKind.Utc);
            return epoch.AddSeconds(seconds);
        }

    }
}

